from types import ModuleType
from typing import Any, Callable, List, Optional, Union
from .grading import test_case, LearnerSubmission

grading_function = Callable[[Any], List[test_case]]
learner_submission = Union[ModuleType, LearnerSubmission]
grading_wrapper = Callable[[learner_submission, Optional[ModuleType]], grading_function]



import argparse
from .generator import init
from .io import update_grader_and_notebook_version, update_notebook_version, tag_notebook
from .config import Config


def parse_dlai_grader_args() -> None:
    parser = argparse.ArgumentParser(
        description="Helper library to build automatic graders for DLAI courses."
    )
    parser.add_argument(
        "-i",
        "--init",
        action="store_true",
        help="Initialize a grader workspace with common files and directories.",
    )
    parser.add_argument(
        "-u",
        "--upgrade",
        action="store_true",
        help="Upgrade the grader and notebook version.",
    )
    parser.add_argument(
        "-v",
        "--versioning",
        action="store_true",
        help="Add version to notebook metadata that matches current grader version.",
    )
    parser.add_argument(
        "-t",
        "--tag",
        action="store_true",
        help="Add graded tag to all code cells of notebook.",
    )
    args = parser.parse_args()

    if args.init:
        init()
    if args.upgrade:
        update_grader_and_notebook_version()
    if args.versioning:
        c = Config()
        update_notebook_version("./mount/submission.ipynb", c.latest_version)
    if args.tag:
        tag_notebook("./mount/submission.ipynb")




import os
import json
import shutil
import tarfile
import nbformat
import jupytext
from os import devnull
from zipfile import ZipFile
from nbformat.notebooknode import NotebookNode
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from .notebook import tag_code_cells
from .config import Config


@contextmanager
def suppress_stdout_stderr():
    """A context manager that redirects stdout and stderr to devnull"""
    with open(devnull, "w") as fnull:
        with redirect_stderr(fnull) as err, redirect_stdout(fnull) as out:
            yield (err, out)


def read_notebook(
    path: str,
) -> NotebookNode:
    """Reads a notebook found in the given path and returns a serialized version.
    Args:
        path (str): Path of the notebook file to read.
    Returns:
        NotebookNode: Representation of the notebook following nbformat convention.
    """
    return nbformat.read(path, as_version=nbformat.NO_CONVERT)


def tag_notebook(
    path: str,
) -> None:
    """Adds 'graded' tag to all code cells of a notebook.

    Args:
        path (str): Path to the notebook.
    """
    nb = read_notebook(path)
    nb = tag_code_cells(nb)
    jupytext.write(nb, path)


def extract_tar(
    file_path: str,
    destination: str,
    post_cleanup: bool = True,
) -> None:
    """Extracts a tar file unto the desired destination.

    Args:
        file_path (str): Path to tar file.
        destination (str): Path where to save uncompressed files.
        post_cleanup (bool, optional): If true, deletes the compressed tar file. Defaults to True.
    """
    with tarfile.open(file_path, "r") as my_tar:
        my_tar.extractall(destination)

    if post_cleanup and os.path.exists(file_path):
        os.remove(file_path)


def extract_zip(
    file_path: str,
    destination: str,
    post_cleanup: bool = True,
) -> None:
    """Extracts a zip file unto the desired destination.

    Args:
        file_path (str): Path to zip file.
        destination (str): Path where to save uncompressed files.
        post_cleanup (bool, optional): If true, deletes the compressed zip file. Defaults to True.
    """
    with ZipFile(file_path, "r") as zip:
        zip.extractall(destination)

    if post_cleanup and os.path.exists(file_path):
        os.remove(file_path)


def send_feedback(
    score: float,
    msg: str,
    feedback_path: str = "/shared/feedback.json",
    err: bool = False,
) -> None:
    """Sends feedback to the learner.
    Args:
        score (float): Grading score to show on Coursera for the assignment.
        msg (str): Message providing additional feedback.
        feedback_path (str): Path where the json feedback will be saved. Defaults to /shared/feedback.json
        err (bool, optional): True if there was an error while grading. Defaults to False.
    """

    post = {"fractionalScore": score, "feedback": msg}
    print(json.dumps(post))

    with open(feedback_path, "w") as outfile:
        json.dump(post, outfile)

    if err:
        exit(1)

    exit(0)


def copy_submission_to_workdir(
    dir_origin: str = "/shared/submission/",
    dir_destination: str = "./submission/",
    file_name: str = "submission.ipynb",
) -> None:
    """Copies submission file from bind mount directory into working directory.
    Args:
        dir_origin (str): Origin directory.
        dir_destination (str): Target directory.
        file_name (str): Name of the file.
    """

    _from = os.path.join(dir_origin, file_name)
    _to = os.path.join(dir_destination, file_name)
    shutil.copyfile(_from, _to)


def update_grader_version() -> str:
    """Updates the grader version by 1 unit.

    Returns:
        str: New version of the grader.
    """
    with open("./.conf", "r") as f:
        lines = f.readlines()

    new_lines = []
    for l in lines:
        if ("GRADER_VERSION" in l) and (not "TAG_ID" in l):
            _, v = l.split("=")
            num_v = int(v)
            new_v = num_v + 1
            new_l = f"GRADER_VERSION={new_v}\n"
            new_lines.append(new_l)
            continue
        new_lines.append(l)

    with open("./.conf", "w") as f:
        f.writelines(new_lines)

    return str(new_v)


def update_notebook_version(
    path: str,
    version: str,
) -> None:
    """Updates notebook version to match the latest version of the grader.

    Args:
        path (str): Path to the notebook.
        version (str): Latest version of the grader to update the notebook to.
    """
    nb = read_notebook(path)
    metadata = nb.get("metadata")
    metadata.update({"dlai_version": version})
    nb["metadata"] = metadata
    jupytext.write(nb, path)


def update_grader_and_notebook_version() -> None:
    """Updates the notebook and the grader at the same time."""
    latest_version = update_grader_version()
    update_notebook_version("./mount/submission.ipynb", latest_version)



import os
from .templates import load_templates


def write_file(filename, template):
    with open(filename, "w") as f:
        f.write(template)


def init():

    template_dict = load_templates()
    write_file("./Dockerfile", template_dict.get("dockerfile"))
    write_file("./grader.py", template_dict.get("grader_py"))
    write_file("./Makefile", template_dict.get("makefile"))
    write_file("./.conf", template_dict.get("conf"))
    write_file("./entry.py", "")
    write_file("./requirements.txt", "dlai-grader")
    os.makedirs("data")
    os.makedirs("learner")
    os.makedirs("mount")
    os.makedirs("solution")
    os.makedirs("submission")



from re import template
from textwrap import dedent
from typing import Dict


def load_templates() -> Dict[str, str]:
    specialization = input("Name of the specialization: ")
    course = input("Number of the course: ")
    week = input("Number of the week: ")
    version = input("Version of the grader (leave empty for version 1): ")
    version = "1" if not version else version

    dockerfile = """
	FROM frolvlad/alpine-miniconda3:python3.7

	RUN apk update && apk add libstdc++

	COPY requirements.txt .

	RUN pip install -r requirements.txt && \
	rm requirements.txt

	RUN mkdir /grader && \ 
	mkdir /grader/submission

	COPY .conf /grader/.conf
	COPY data/ /grader/data/
	COPY solution/ /grader/solution/
	COPY entry.py /grader/entry.py
	COPY grader.py /grader/grader.py

	RUN chmod a+rwx /grader/

	WORKDIR /grader/

	ENTRYPOINT ["python", "entry.py"]
    """

    conf = f"""
	ASSIGNMENT_NAME=C{course}W{week}_Assignment
	IMAGE_NAME={specialization}c{course}w{week}-grader
	GRADER_VERSION={version}
	TAG_ID=V$(GRADER_VERSION)
	SUB_DIR=mount
	MEMORY_LIMIT=4096
	LINUX_UPLOAD_DIR='/mnt/c/Users/Andres/dlai/upload'
	MAC_UPLOAD_DIR='/Users/andreszarta/Desktop/upload-temp'
    """

    makefile = """
	include .conf

	PARTIDS = ""
	OS := $(shell uname)

	learner:
		[ "$(ls -A ./$(SUB_DIR))" ] && rm ./$(SUB_DIR)/* && echo "Deleted previous version"|| echo "Empty"
		cp mount/submission.ipynb learner/$(ASSIGNMENT_NAME)_Solution.ipynb
		learner -fn learner/$(ASSIGNMENT_NAME)_Solution.ipynb 
		rm learner/$(ASSIGNMENT_NAME)_Solution.ipynb && rm learner/$(ASSIGNMENT_NAME)_Solution.py && rm learner/$(ASSIGNMENT_NAME).py

	build:
		docker build -t $(IMAGE_NAME):$(TAG_ID) .

	entry:
		docker run -it --rm --mount type=bind,source=$(PWD)/mount,target=/shared/submission --mount type=bind,source=$(PWD),target=/grader/ --entrypoint ash $(IMAGE_NAME):$(TAG_ID)

	test:
		docker run -it --rm --mount type=bind,source=$(PWD)/mount,target=/shared/submission --mount type=bind,source=$(PWD),target=/grader/ --entrypoint pytest $(IMAGE_NAME):$(TAG_ID)

	grade:
		gradethis $(PARTIDS) $(IMAGE_NAME):$(TAG_ID) $(SUB_DIR) $(MEMORY_LIMIT)

	mem:
		memthis $(PARTIDS)

	zip:
		zip -r $(IMAGE_NAME)$(TAG_ID).zip .

	clean:
		find . -maxdepth 1 -type f -name "*.zip" -exec rm {} +
		docker rm $$(docker ps -qa --no-trunc --filter "status=exited")
		docker rmi $$(docker images --filter "dangling=true" -q --no-trunc)

	upload:
		coursera_autograder --timeout 1800 upload --grader-memory-limit $(MEMORY_LIMIT) --grading-timeout 1800 $(IMAGE_NAME)$(TAG_ID).zip $(COURSE_ID) $(ITEM_ID) $(PART_ID)

	move-zip:
		if [[ "$(OS)" == "Darwin" ]];    \
		then    \
			mv $(IMAGE_NAME)$(TAG_ID).zip $(MAC_UPLOAD_DIR);    \
		else    \
			mv $(IMAGE_NAME)$(TAG_ID).zip $(LINUX_UPLOAD_DIR);    \
		fi

	move-learner:
		if [[ "$(OS)" == "Darwin" ]];    \
		then    \
			cp learner/$(ASSIGNMENT_NAME).ipynb $(MAC_UPLOAD_DIR);    \
		else    \
			cp learner/$(ASSIGNMENT_NAME).ipynb $(LINUX_UPLOAD_DIR);    \
		fi
    """


    grader_py = """
	from types import ModuleType, FunctionType, FunctionType
	from typing import Dict, List, Optional
	from dlai_grader.grading import test_case, object_to_grade
	from dlai_grader.types import grading_function, grading_wrapper


	def part_1(
		learner_mod: ModuleType, solution_mod: Optional[ModuleType]
	) -> grading_function:
		@object_to_grade(learner_mod, "learner_func")
		def g(learner_func: FunctionType) -> List[test_case]:

			t = test_case()
			if not isinstance(learner_func, FunctionType):
				t.failed = True
				t.msg = "learner_func has incorrect type"
				t.want = FunctionType
				t.got = type(learner_func)
				return [t]

			cases: List[test_case] = []

			return cases

		return g


	def handle_part_id(part_id: str) -> grading_wrapper:
		grader_dict: Dict[str, grading_wrapper] = {
			"": part_1,
		}
		return grader_dict[part_id]
	"""


    template_dict = {
		"dockerfile": dedent(dockerfile[1:]),
		"makefile": dedent(makefile[1:]),
		"conf": dedent(conf[1:]),
		"grader_py": dedent(grader_py[1:])
	}

    return template_dict



from dataclasses import dataclass
from functools import wraps
from types import ModuleType
from typing import Any, Callable, List, Tuple


@dataclass
class LearnerSubmission:
    """Class that represents a file from the learner. Useful when grading does not depend on the notebook"""

    submission: Any = None


@dataclass
class test_case:
    """Class that represents a test case"""

    msg: str = ""
    want: Any = None
    got: Any = None
    failed: bool = False


@dataclass
class aggregated_test_case:
    """Class that represents an aggregated collection of test cases"""

    test_name: str
    tests: List[test_case]


def compute_grading_score(test_cases: List[test_case],) -> Tuple[float, str]:
    """Computes the score based on the number of failed and total cases.
    Args:
        test_cases (List): Test cases.
    Returns:
        Tuple[float, str]: The grade and feedback message.
    """

    num_cases = len(test_cases)
    failed_cases = [t for t in test_cases if t.failed == True]
    score = 1.0 - len(failed_cases) / num_cases
    feedback_msg = "All tests passed! Congratulations!"

    if failed_cases:
        feedback_msg = ""
        for failed_case in failed_cases:
            feedback_msg += f"Failed test case: {failed_case.msg}.\nExpected:\n{failed_case.want},\nbut got:\n{failed_case.got}.\n\n"
        return round(score, 2), feedback_msg

    return score, feedback_msg


def compute_aggregated_grading_score(
    aggregated_test_cases: List[aggregated_test_case],
) -> Tuple[float, str]:
    """Computes the score based on the number of failed and total cases.
    Args:
        aggregated_test_cases (List): Aggregated test cases for every part.
    Returns:
        Tuple[float, str]: The grade and feedback message.
    """
    scores = []
    msgs = []

    for test_cases in aggregated_test_cases:
        feedback_msg = f"All tests passed for {test_cases.test_name}!\n"
        num_cases = len(test_cases.tests)
        failed_cases = [t for t in test_cases.tests if t.failed == True]
        score = 1.0 - len(failed_cases) / num_cases
        score = round(score, 2)
        scores.append(score)

        if failed_cases:
            feedback_msg = f"Details of failed tests for {test_cases.test_name}\n\n"
            for failed_case in failed_cases:
                feedback_msg += f"Failed test case: {failed_case.msg}.\nExpected:\n{failed_case.want},\nbut got:\n{failed_case.got}.\n\n"
        msgs.append(feedback_msg)

    final_score = sum(scores) / len(scores)
    final_score = round(final_score, 2)
    final_msg = "\n".join(msgs)

    return final_score, final_msg


def object_to_grade(
    origin_module: ModuleType, attr_name: str,
) -> Callable[[Callable[[Any], List[test_case]]], Callable[[Any], List[test_case]]]:
    """Used as a parameterized decorator to get an attribute from a module.

    Args:
        origin_module (ModuleType): A module.
        attr_name (str): Name of the attribute to extract from the module.
    """

    def middle(func):
        @wraps(func)
        def wrapper():
            val = getattr(origin_module, attr_name, None)
            return func(val)

        return wrapper

    return middle



import re
from typing import Callable
import jupytext
from nbformat.notebooknode import NotebookNode
from .config import Config


def notebook_to_script(
    notebook: NotebookNode,
) -> str:
    """Converts a notebook into a python script serialized as a string.
    Args:
        notebook (NotebookNode): Notebook to convert into script.
    Returns:
        str: Python script representation as string.
    """
    return jupytext.writes(notebook, fmt="py:percent")


def cut_notebook(
    regex_pattern: str = "(grade)(.|[ \t]*)(up)(.|[ \t]*)(to)(.|[ \t]*)(here)",
) -> NotebookNode:
    """Cuts a notebook, this allows for partial grading. Written as a closure so it can be consumed as a functional option for notebooks.
    Args:
        regex_pattern (str): Regexp pattern to look for. Cells after match will be omitted.
    Returns:
        Callable[[NotebookNode], NotebookNode]: A function that cuts the notebook.
    """

    def inner(
        notebook: NotebookNode,
    ):
        """Cuts a notebook by excluding all cells after a pattern is matched.
        Args:
            notebook (NotebookNode): Notebook to filter.
        Returns:
            NotebookNode: The filtered notebook.
        """
        filtered_cells = []

        for cell in notebook["cells"]:
            filtered_cells.append(cell)

            if cell["cell_type"] == "code" and re.search(regex_pattern, cell["source"]):
                break

        notebook["cells"] = filtered_cells
        return notebook

    return inner


def keep_tagged_cells(
    tag: str = "graded",
) -> Callable[[NotebookNode], NotebookNode]:
    """Keeps tagged cells from a notebook. Written as a closure so it can be consumed as a functional option for notebooks.
    Args:
        tag (str): Tag to look for within cell's metadata. Defaults to "graded".
    Returns:
        Callable[[NotebookNode], NotebookNode]: A function that filters the notebook.
    """

    def inner(
        notebook: NotebookNode,
    ) -> NotebookNode:
        """Filters a notebook by including tagged cells.
        Args:
            notebook (NotebookNode): Notebook to filter.
        Returns:
            NotebookNode: The notebook with tagged cells.
        """
        filtered_cells = []

        for cell in notebook["cells"]:
            if not tag in cell["metadata"].get("tags"):
                continue
            filtered_cells.append(cell)

        notebook["cells"] = filtered_cells
        return notebook

    return inner


def omit_tagged_cells(
    tag: str = "omit",
) -> Callable[[NotebookNode], NotebookNode]:
    """Omits tagged cells from a notebook. Written as a closure so it can be consumed as a functional option for notebooks.
    Args:
        tag (str): Tag to look for within cell's metadata. Defaults to "omit".
    Returns:
        Callable[[NotebookNode], NotebookNode]: A function that filters the notebook.
    """

    def inner(
        notebook: NotebookNode,
    ) -> NotebookNode:
        """Filters a notebook by excluding tagged cells.
        Args:
            notebook (NotebookNode): Notebook to filter.
        Returns:
            NotebookNode: The notebook without omitted cells.
        """
        filtered_cells = []

        for cell in notebook["cells"]:
            if tag in cell["metadata"].get("tags"):
                continue
            filtered_cells.append(cell)

        notebook["cells"] = filtered_cells
        return notebook

    return inner


def get_named_cells(
    notebook: NotebookNode,
) -> dict:
    """Returns the named cells for cases when grading is done using cell's output.
    Args:
        notebook (NotebookNode): The notebook from the learner.
    Returns:
        dict: All named cells encoded as a dictionary.
    """
    named_cells = {}
    for cell in notebook["cells"]:
        metadata = cell["metadata"]
        if not "name" in metadata:
            continue
        named_cells.update({metadata.get("name"): cell})
    return named_cells


def tag_code_cells(
    notebook: NotebookNode,
    tag: str = "graded",
) -> NotebookNode:
    """Filters a notebook to exclude additional cells created by learners.
       Also used for partial grading if the tag has been provided.
    Args:
        notebook (NotebookNode): Notebook to filter.
        tag (str): The tag to include in the code cell's metadata. Defaults to "graded".
    Returns:
        NotebookNode: The filtered notebook.
    """
    filtered_cells = []

    for cell in notebook["cells"]:

        if cell["cell_type"] == "code":

            if not "tags" in cell["metadata"]:
                cell["metadata"]["tags"] = []

            tags = cell["metadata"]["tags"]

            if not tag in tags:
                tags.append(tag)
                cell["metadata"]["tags"] = tags

        filtered_cells.append(cell)

    notebook["cells"] = filtered_cells

    return notebook


def notebook_version(
    notebook: NotebookNode,
) -> str:
    """Returns dlai version of a notebook.

    Args:
        notebook (NotebookNode): A notebook.

    Returns:
        str: Version encoded as string.
    """
    return notebook.get("metadata").get("dlai_version")


def notebook_is_up_to_date(
    notebook: NotebookNode,
) -> bool:
    """Determines if a notebook is up-to-date with latest grader version.

    Args:
        notebook (NotebookNode): A notebook.

    Returns:
        bool: True if both versions match, False otherwise.
    """
    version = notebook_version(notebook)
    c = Config()
    if version != c.latest_version:
        return False
    return True
